import itertools
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from hyfi.composer import BaseModel

logger = logging.getLogger(__name__)


class LdaRunConfig(BaseModel):
    _config_group_ = "/runner/config"
    _config_name_ = "lda"

    k: List[int] = [10, 20]
    alpha: List[float] = [0.1]
    eta: List[float] = [0.01]

    @property
    def total_runs(self) -> int:
        return len(self.k) * len(self.alpha) * len(self.eta)

    def iter_configs(self) -> Dict[str, Any]:
        for k, alpha, eta in itertools.product(self.k, self.alpha, self.eta):
            yield dict(k=k, alpha=alpha, eta=eta)


class TopicRunnerResult(BaseModel):
    runner_task_name: str = "topic"
    runner_batch_name: str = "runner"
    runner_batch_num: int = 0
    runner_batch_id: str = ""
    overrides: List[Dict] = []
    model_summary: List[Dict] = []

    def add_model_summary(
        self,
        overrides: Dict,
        summary: Dict,
    ):
        self.overrides.append(overrides)
        self.model_summary.append(summary)


class InferConfig(BaseModel):
    _config_group_ = "/runner/config"
    _config_name_ = "infer_topics"

    model_config_file: Optional[Union[str, Path]] = None
    output_file: Optional[Union[str, Path]] = None
    iterations: int = 100
    tolerance: float = -1
    num_workers: int = 0
    together: bool = False



import logging
from typing import Dict, List, Optional, Union

from hyfi import HyFI
from hyfi.composer import BaseModel

logger = logging.getLogger(__name__)

PriorType = Dict[int, List[str]]


class WordPrior(BaseModel):
    _config_group_: str = "/words"
    _config_name_: str = "wordprior"

    lowercase: bool = True
    prior_data: Optional[PriorType] = None
    data_file: Optional[str] = None
    min_prior_weight: float = 0.01
    max_prior_weight: float = 1.0
    verbose: bool = False

    _priors_: PriorType = {}

    @property
    def priors(self) -> List[str]:
        if not self._priors_:
            self.load()
        return self._priors_

    def add(self, id_: int, prior: Union[str, List[str]]):
        """Add a prior to the word prior"""
        if isinstance(prior, str):
            prior = [prior]
        prior = [w.lower() for w in prior] if self.lowercase else prior
        self._priors_[id_] = list(set(self._priors_.get(id_, []) + prior))

    def load(self):
        """Load words from file or function"""
        if self.data_file:
            self._priors_ = HyFI.to_dict(HyFI.load(self.data_file))
            if self.verbose:
                logger.info(
                    "Loaded %d words from %s",
                    len(self._priors_),
                    self.data_file,
                )
        else:
            self._priors_ = {}

        if self.prior_data:
            for id_, prior in self.prior_data.items():
                self._priors_[id_] = prior

        self._priors_ = {
            int(id_): list({w.lower() if self.lowercase else w for w in prior})
            for id_, prior in self._priors_.items()
        }
        if self.verbose:
            logger.info("Loaded %d priors", len(self._priors_))

    def save(self):
        """Save words to file"""
        if not self.priors:
            logger.warning("No priors to save")
            return
        HyFI.save(self.priors, self.data_file)
        if self.verbose:
            logger.info("Saved %d words to %s", len(self.priors), self.data_file)

    def items(self):
        return self.priors.items()

    def __iter__(self):
        return iter(self.priors)

    def __len__(self):
        return len(self.priors)

    def __repr__(self):
        return f"<WordPrior {len(self.priors)} priors>"

    def __str__(self):
        return f"<WordPrior {len(self.priors)} priors>"

    def __bool__(self):
        return bool(self.priors)

    def __eq__(self, other: "WordPrior"):
        return self.priors == other.priors

    def __ne__(self, other: "WordPrior"):
        return self.priors != other.priors



from .base import TopicModel
from .config import LdaConfig, TrainConfig, TrainSummaryConfig, WordcloudConfig
from .lda import LdaModel
from .prior import WordPrior

__all__ = [
    "LdaModel",
    "LdaConfig",
    "TopicModel",
    "WordPrior",
    "TrainConfig",
    "TrainSummaryConfig",
    "WordcloudConfig",
]



import logging
from pathlib import Path
from typing import Callable, Dict, Optional, Set, Tuple, Union

import numpy as np
import wordcloud
from hyfi import HyFI
from hyfi.composer import BaseModel

logger = logging.getLogger(__name__)


class WordCloud(BaseModel):
    _config_group_ = "/plot"
    _config_name_ = "wordcloud"

    font_path: Optional[str] = None
    width: int = 400
    height: int = 200
    prefer_horizontal: float = 0.9
    mask: Optional[Union[str, np.ndarray]] = None
    contour_width: float = 0
    contour_color: str = "steelblue"
    scale: float = 1
    min_font_size: int = 4
    max_font_size: Optional[int] = None
    font_step: int = 1
    max_words: int = 200
    stopwords: Optional[Set] = None
    background_color: str = "black"
    mode: str = "RGB"
    relative_scaling: Union[float, str] = "auto"
    color_func: Optional[Callable] = None
    regexp: Optional[str] = None
    collocations: bool = True
    colormap: str = "PuBu"
    normalize_plurals: bool = True
    repeat: bool = False
    include_numbers: bool = False
    min_word_length: int = 0
    collocation_threshold: int = 30

    _wc_: Optional[wordcloud.WordCloud] = None

    @property
    def wc(self) -> Optional[wordcloud.WordCloud]:
        return self._wc_

    def generate_from_frequencies(
        self,
        frequencies: Dict[str, float],
        max_font_size: Optional[int] = None,
        output_file: Optional[str] = None,
        lang: str = "en",
        verbose: bool = False,
    ) -> wordcloud.WordCloud:
        _, self.font_path = HyFI.get_plot_font(
            fontpath=self.font_path,
            lang=lang,
        )
        if max_font_size:
            self.max_font_size = max_font_size

        if isinstance(self.mask, str) and HyFI.is_file(self.mask):
            if verbose:
                logger.info("Loading mask from %s", self.mask)
            self.mask = HyFI.load_image_as_ndarray(self.mask)
            self.width = self.mask.shape[1]
            self.height = self.mask.shape[0]

        self._wc_ = wordcloud.WordCloud(**self.kwargs)
        self._wc_.generate_from_frequencies(frequencies)
        if output_file:
            Path(output_file).parent.mkdir(parents=True, exist_ok=True)
            self._wc_.to_file(output_file)
            if verbose:
                logger.debug("Saved wordcloud to %s", output_file)
        return self._wc_



import logging
from pathlib import Path
from typing import Dict, List, Optional, Union

from hyfi import HyFI
from hyfi.runner import BaseRunner
from tqdm.auto import tqdm

from thematos.datasets import Corpus
from thematos.models import LdaModel

from .config import InferConfig, LdaRunConfig, TopicRunnerResult

logger = logging.getLogger(__name__)


class TopicRunner(BaseRunner):
    _config_group_ = "/runner"
    _config_name_ = "topic"

    task_name: str = "topic"
    batch_name: str = "runner"
    model: LdaModel = LdaModel()
    run_args: LdaRunConfig = LdaRunConfig()
    corpus_to_infer: Corpus = Corpus()
    infer_args: InferConfig = InferConfig()

    num_workers: int = 0
    verbose: bool = False

    calls: Optional[List[Union[str, Dict]]] = ["train"]

    _summaries_: Optional[TopicRunnerResult] = None

    def load_model(
        self,
        batch_name: Optional[str] = None,
        batch_num: Optional[int] = None,
        filepath: Optional[Union[str, Path]] = None,
        **config_kwargs,
    ) -> None:
        self.model.load(
            batch_name=batch_name,
            batch_num=batch_num,
            filepath=filepath,
            **config_kwargs,
        )

    def infer(self) -> None:
        if self.verbose:
            logger.info("Running inference with args: %s", self.infer_args)
        self.load_model(filepath=self.infer_args.model_config_file)
        self.model.infer(
            corpus=self.corpus_to_infer,
            output_file=self.infer_args.output_file,
            iterations=self.infer_args.iterations,
            tolerance=self.infer_args.tolerance,
            num_workers=self.infer_args.num_workers,
            together=self.infer_args.together,
        )

    def train(self) -> None:
        self._summaries_ = TopicRunnerResult(
            runner_task_name=self.task_name,
            runner_batch_name=self.batch_name,
            runner_batch_num=self.batch_num,
            runner_batch_id=self.batch_id,
        )
        self.model.batch.num_workers = self.num_workers
        for args in tqdm(self.run_args.iter_configs(), total=self.run_args.total_runs):
            if self.verbose:
                logger.info("Running with args: %s", args)
            self.model.update_model_args(**args)
            self.model.train()
            self._summaries_.add_model_summary(
                overrides=args,
                summary=self.model.model_summary_dict,
            )
            self.model.set_batch_num(self.model.batch_num + 1)
        self.save()

    def save(self) -> None:
        self.save_result_summary()
        self.save_config()

    @property
    def result_summary_file(self) -> Path:
        summary_file = f"{self.batch_id}_summaries.json"
        return self.task_dir / summary_file

    @property
    def summaries(self) -> Dict:
        return self._summaries_.model_dump() if self._summaries_ else {}

    def save_result_summary(self) -> None:
        if self._summaries_:
            HyFI.save_json(self.summaries, self.result_summary_file)
            logger.info("Saved summaries to %s", self.result_summary_file)
        else:
            logger.warning("No summaries to save")



from typing import Any, Dict, Optional

import tomotopy as tp
from hyfi.composer import BaseModel


class CoherenceMetrics(BaseModel):
    u_mass: Optional[float] = None
    c_uci: Optional[float] = None
    c_npmi: Optional[float] = None
    c_v: Optional[float] = None


class ModelSummary(BaseModel):
    timestamp: str = None
    model_id: str = None
    model_type: str = None
    num_docs: int = None
    num_words: int = None
    num_total_vocabs: Optional[int] = None
    num_used_vocabs: int = None
    seed: Optional[int] = None
    model_args: Dict[str, Any] = None
    train_args: Dict[str, Any] = None
    perplexity: float = None
    coherence: Optional[Dict[str, float]] = None


IDF = tp.TermWeight.IDF
ONE = tp.TermWeight.ONE
PMI = tp.TermWeight.PMI



import sys
from typing import Any, List, Optional, Tuple

from hyfi.composer import BaseModel, Field

from thematos.plots import WordCloud

from .types import IDF


class LdaConfig(BaseModel):
    _config_group_ = "/model/config"
    _config_name_ = "lda"

    tw: int = int(IDF.value)
    min_cf: int = 5
    min_df: int = 0
    rm_top: int = 0
    k: Optional[int] = None
    alpha: float = 0.1
    eta: float = 0.01


class TrainConfig(BaseModel):
    _config_group_ = "/model/train"
    _config_name_ = "topic"

    burn_in: int = 0
    interval: int = 10
    iterations: int = 100


class TrainSummaryConfig(BaseModel):
    _config_group_ = "/model/summary"
    _config_name_ = "topic_train"

    initial_hp: bool = Field(
        True, description="whether to show the initial parameters at model creation"
    )
    params: bool = Field(
        True, description="whether to show the current parameters of the model"
    )
    topic_word_top_n: int = Field(
        10, description="the number of words by topic to display"
    )
    flush: bool = Field(False, description="whether to forcibly flush the stream")


class WordcloudConfig(BaseModel):
    _config_group_ = "/model/plot"
    _config_name_ = "wordcloud"

    wc: WordCloud = WordCloud()
    top_n: int = 100
    make_collage: bool = True
    num_images_per_page: int = 20
    num_cols: int = 5
    num_rows: Optional[int] = None
    output_file_format: str = "wordcloud_p{page_num:02d}.png"
    titles: Optional[List[str]] = None
    title_fontsize: int = 14
    title_color: str = "green"
    figsize: Optional[Tuple[int, int]] = None
    width_multiple: float = 4
    height_multiple: float = 2
    dpi: int = 300
    mask_dir: Optional[str] = None
    fontpath: Optional[str] = None
    save: bool = True



import logging
from pathlib import Path
from typing import Any, List, Optional

import numpy as np
import pandas as pd
import tomotopy as tp
from hyfi import HyFI
from hyfi.run import RunConfig
from hyfi.task import BatchTaskConfig
from lexikanon.stopwords import Stopwords

from .ngrams import NgramConfig

logger = logging.getLogger(__name__)


class Corpus(BatchTaskConfig):
    _config_group_: str = "/dataset"
    _config_name_: str = "topic_corpus"

    task_name: str = "topic"
    batch_name: str = "corpus"
    id_col: str = "id"
    text_col: str = "text"
    timestamp_col: Optional[str] = None
    data_load: RunConfig = RunConfig(_config_name_="load_dataframe")
    stopwords: Optional[Stopwords] = Stopwords()
    ngrams: Optional[NgramConfig] = NgramConfig()
    ngramize: bool = True
    verbose: bool = False

    _data_: Optional[pd.DataFrame] = None
    _corpus_: Optional[tp.utils.Corpus] = None
    _doc_ids_: List[Any] = None

    @property
    def data(self) -> pd.DataFrame:
        if self._data_ is None:
            self._data_ = self.load_data()
            if self.verbose:
                logger.info(self._data_[self.text_col].head())
        return self._data_

    def load_data(self) -> pd.DataFrame:
        return HyFI.partial(self.data_load.config)()

    @property
    def corpus(self) -> tp.utils.Corpus:
        if self._corpus_ is None:
            self.load()
        return self._corpus_

    def load(self):
        if self._corpus_:
            logger.infro("Corpus is already loaded")
            return
        logger.info("Loading corpus...")
        self._doc_ids_ = []
        self._corpus_ = tp.utils.Corpus(
            tokenizer=tp.utils.SimpleTokenizer(),
            stopwords=self.stopwords,
        )
        logger.info("Processing documents in the column '%s'...", self.text_col)
        self._corpus_.process(self.docs)
        logger.info("Total %d documents are loaded.", len(self.docs))
        if self.ngramize:
            self.concat_ngrams()
        self.save_ids()
        self.save_config()

    @property
    def docs(self) -> List[str]:
        doc = self.data[self.text_col].head(1).values[0]
        if self.verbose:
            logger.info("The first document: %s", doc)
        # check the type of the first document to see if it is a list of words
        # or a string
        # if the type is ndarray, convert it to string
        if isinstance(doc, str):
            return self.data[self.text_col].values.tolist()
        elif isinstance(doc, np.ndarray or list):
            return self.data[self.text_col].str.join(" ").values.tolist()

    @property
    def doc_ids(self) -> List[Any]:
        if not self._doc_ids_:
            self._doc_ids_ = self.data[self.id_col].values.tolist()
        return self._doc_ids_

    @property
    def id_file(self) -> Path:
        return self.batch_dir / f"{self.batch_name}_doc_ids.parquet"

    def save_ids(self):
        """Save doc ids to file"""
        data = pd.DataFrame(self.doc_ids, columns=[self.id_col])
        HyFI.save_dataframes(data, self.id_file, verbose=self.verbose)

    def __repr__(self):
        if self._doc_ids_:
            return f"<Corpus {len(self._doc_ids_)} documents>"
        else:
            return "<Corpus>"

    def __str__(self):
        if self._doc_ids_:
            return f"<Corpus {len(self._doc_ids_)} documents>"
        else:
            return "<Corpus>"

    def __bool__(self):
        return bool(self._doc_ids_)

    def concat_ngrams(self, delimiter: Optional[str] = None):
        if not self.ngrams:
            logger.info("N-grams config is not set. Skipping...")
            return
        # extract the n-gram candidates first
        if not delimiter:
            delimiter = self.ngrams.delimiter
        args = self.ngrams.kwargs
        args.pop("delimiter", None)
        if not self.ngrams.min_score:
            args.pop("min_score", None)
        if self.verbose:
            logger.info("Extracting n-grams...")
        cands = self.corpus.extract_ngrams(**args)
        if self.verbose:
            logger.info("Total %d n-grams are extracted.", len(cands))
            for cand in cands[:10]:
                logger.info(cand)
            for cand in cands[-10:]:
                logger.info(cand)
        # concat n-grams in the corpus
        if self.verbose:
            logger.info("Concatenating n-grams...")
            logger.info(self.corpus[0])
            logger.info(self.corpus[-1])
        self.corpus.concat_ngrams(cands, delimiter=delimiter)
        if self.verbose:
            logger.info(self.corpus[0])
            logger.info(self.corpus[-1])
            logger.info("Total %d documents are n-gramized.", len(self.corpus))



import logging
from typing import Optional

import tomotopy as tp
from hyfi import HyFI
from tqdm.auto import tqdm

from .base import TopicModel
from .config import LdaConfig

logger = logging.getLogger(__name__)


class LdaModel(TopicModel):
    _config_name_ = "lda"

    model_type: str = "LDA"
    model_args: LdaConfig = LdaConfig()

    # internal attributes
    _model_: Optional[tp.LDAModel] = None

    @property
    def model(self) -> tp.LDAModel:
        if self._model_ is None:
            self._model_ = tp.LDAModel(
                corpus=self.tp_corpus,
                seed=self.seed,
                **self.model_args_dict,
            )
        return self._model_

    def _train(self, model: tp.LDAModel) -> None:
        train_args = self.train_args
        model.burn_in = train_args.burn_in
        model.train(0)
        logger.info("Number of docs: %s", len(model.docs))
        logger.info("Vocab size: %s", model.num_vocabs)
        logger.info("Number of words: %s", model.num_words)
        logger.info("Removed top words: %s", model.removed_top_words)
        logger.info(
            "Training model by iterating over the corpus %s times, %s iterations at a time with %s workers",
            train_args.iterations,
            train_args.interval,
            self.batch.num_workers,
        )

        ll_per_words = []
        for i in tqdm(range(0, train_args.iterations, train_args.interval)):
            model.train(
                iter=train_args.interval,
                workers=self.batch.num_workers,
            )
            logger.info("Iteration: %s\tLog-likelihood: %s", i, model.ll_per_word)
            ll_per_words.append((i, model.ll_per_word))
        self._ll_per_words_ = ll_per_words
        if self.verbose:
            self.model.summary(**self.train_summary_args.kwargs)
        self._model_ = model

    def _load_model(self):
        model_path = self.model_file
        if HyFI.is_file(model_path):
            self._model_ = tp.LDAModel.load(model_path)
            logger.info("Model loaded from %s", model_path)
        else:
            self._model_ = None
            logger.warning("Model file %s does not exist", model_path)



import logging
import sys
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import tomotopy as tp
from hyfi import HyFI
from hyfi.task import BatchTaskConfig

from thematos.datasets import Corpus

from .config import LdaConfig, TrainConfig, TrainSummaryConfig, WordcloudConfig
from .prior import WordPrior
from .types import CoherenceMetrics, ModelSummary

logger = logging.getLogger(__name__)


class TopicModel(BatchTaskConfig):
    _config_group_ = "/model"
    _config_name_ = "topic"

    task_name: str = "topic"
    batch_name: str = "model"
    model_type: str = "BASE"
    wordprior: WordPrior = WordPrior()
    corpus: Corpus = Corpus()
    model_args: LdaConfig = LdaConfig()
    train_args: TrainConfig = TrainConfig()
    train_summary_args: TrainSummaryConfig = TrainSummaryConfig()
    wc_args: WordcloudConfig = WordcloudConfig()

    coherence_metric_list: List[str] = ["u_mass", "c_uci", "c_npmi", "c_v"]
    eval_coherence: bool = True
    set_wordprior: bool = False
    autosave: bool = True
    save_full: bool = True
    verbose: bool = False

    # internal attributes
    _model_: Optional[Any] = None
    _timestamp_: Optional[str] = None
    _coherence_metrics_: Optional[CoherenceMetrics] = None
    _model_summary_: Optional[ModelSummary] = None
    _ll_per_words_: List[Tuple[int, float]] = []
    _doc_ids_: List[Any] = None
    _doc_topic_dists_df_: Optional[pd.DataFrame] = None
    _topic_term_dists_df_: Optional[pd.DataFrame] = None

    @property
    def model_id(self) -> str:
        model_type = self.model_type.upper()
        margs = [model_type, self.batch_id, f"k({self.model_args.k})"]
        return "_".join(margs)

    @property
    def model(self):
        if self._model_ is None:
            raise ValueError("Model has not been trained yet.")
        return self._model_

    @property
    def coherence_metrics_dict(self) -> Dict:
        return self._coherence_metrics_.model_dump() if self._coherence_metrics_ else {}

    @property
    def model_summary_dict(self) -> Dict:
        return self._model_summary_.model_dump() if self._model_summary_ else {}

    @property
    def train_args_dict(self) -> Dict:
        return self.train_args.kwargs if self.train_args else {}

    @property
    def model_args_dict(self) -> Dict:
        return self.model_args.kwargs if self.model_args else {}

    @property
    def timestamp(self) -> str:
        if self._timestamp_ is None:
            raise ValueError("Model has not been trained yet.")
        return self._timestamp_

    @property
    def tp_corpus(self) -> tp.utils.Corpus:
        return self.corpus.corpus

    @property
    def doc_ids(self) -> List[Any]:
        if not self._doc_ids_:
            self._doc_ids_ = self.corpus.doc_ids
        return self._doc_ids_

    @property
    def ll_per_words(self) -> Optional[pd.DataFrame]:
        if not self._ll_per_words_:
            logger.warning("No log-likelihood per word found.")
            return None
        return pd.DataFrame(self._ll_per_words_, columns=["iter", "ll_per_word"])

    @property
    def doc_topic_dists(self) -> np.ndarray:
        dist_ = np.stack([doc.get_topic_dist() for doc in self.model.docs])
        dist_ /= dist_.sum(axis=1, keepdims=True)
        return dist_

    @property
    def topic_term_dists(self) -> np.ndarray:
        return np.stack(
            [self.model.get_topic_word_dist(k) for k in range(self.model.k)]
        )

    @property
    def doc_num_words(self) -> np.ndarray:
        return np.array([len(doc.words) for doc in self.model.docs])

    @property
    def used_vocab(self) -> List[str]:
        return list(self.model.used_vocabs)

    @property
    def term_frequency(self) -> np.ndarray:
        return self.model.used_vocab_freq

    @property
    def num_docs(self) -> int:
        return len(self.model.docs)

    @property
    def num_words(self) -> int:
        return self.model.num_words

    @property
    def num_total_vocabs(self) -> int:
        return len(self.model.vocabs) if self.model.vocabs else None

    @property
    def num_used_vocab(self) -> int:
        return len(self.model.used_vocabs)

    @property
    def num_topics(self) -> int:
        """Number of topics in the model

        It is the same as the number of columns in the document-topic distribution.
        """
        return self.model.k if self.model else len(self.doc_topic_dists[0])

    @property
    def topic_term_dists_df(self) -> pd.DataFrame:
        return pd.DataFrame(self.topic_term_dists, columns=self.used_vocab)

    def get_doc_topic_dists_df(
        self,
        doc_topic_dists: Optional[np.ndarray],
        doc_ids: Optional[List[Any]],
    ) -> pd.DataFrame:
        if len(doc_topic_dists) != len(doc_ids):
            raise ValueError(
                f"Number of inferred topics ({len(doc_topic_dists)}) does not match with number of documents ({len(doc_ids)})"
            )
        columns = [f"topic{i}" for i in range(self.num_topics)]
        dists_df = pd.DataFrame(doc_topic_dists, columns=columns)
        doc_id_df = pd.DataFrame(doc_ids, columns=["id"])
        return pd.concat([doc_id_df, dists_df], axis=1)

    @property
    def model_file(self) -> str:
        f_ = f"{self.model_id}.mdl"
        self.model_dir.mkdir(parents=True, exist_ok=True)
        return str(self.model_dir / f_)

    @property
    def ll_per_words_file(self) -> str:
        f_ = f"{self.model_id}-ll_per_word.csv"
        return str(self.output_dir / f_)

    @property
    def ll_per_words_fig_file(self) -> str:
        f_ = f"{self.model_id}-ll_per_word.png"
        return str(self.output_dir / f_)

    @property
    def doc_topic_dists_file(self) -> str:
        f_ = f"{self.model_id}-doc_topic_dists.parquet"
        return str(self.output_dir / f_)

    @property
    def topic_term_dists_file(self) -> str:
        f_ = f"{self.model_id}-topic_term_dists.parquet"
        return str(self.output_dir / f_)

    @property
    def used_vocab_file(self) -> str:
        f_ = f"{self.model_id}-used_vocab.txt"
        return str(self.output_dir / f_)

    @property
    def train_summary_file(self) -> str:
        f_ = f"{self.model_id}-summary.txt"
        return str(self.output_dir / f_)

    @property
    def batch_model_summary_file(self) -> str:
        f_ = f"{self.batch_name}-summary.jsonl"
        return str(self.output_dir / f_)

    @property
    def ldavis_file(self) -> str:
        f_ = f"{self.model_id}-ldavis.html"
        return str(self.output_dir / f_)

    @property
    def topic_wordcloud_file_format(self) -> str:
        format_ = self.model_id + "-wordcloud_{topic_id:03d}.png"
        return str(self.output_dir / "wordclouds" / format_)

    def update_model_args(self, **kwargs) -> None:
        self.model_args = self.model_args.model_copy(update=kwargs)

    def _set_wordprior(self) -> None:
        if self.wordprior is None:
            logger.info("No word prior set.")
            return
        if self.verbose:
            logger.info("Set word prior with %s.", self.wordprior)
        for tno, words in self.wordprior.items():
            if self.verbose:
                logger.info("Set words %s to topic #%s as prior.", words, tno)
            for word in words:
                self.model.set_word_prior(
                    word,
                    [
                        self.wordprior.max_prior_weight
                        if i == int(tno)
                        else self.wordprior.min_prior_weight
                        for i in range(self.num_topics)
                    ],
                )

    def train(self) -> None:
        # reset model
        self._model_ = None

        if self.set_wordprior:
            self._set_wordprior()

        self._timestamp_ = datetime.now().strftime("%Y%m%d_%H%M%S")
        # train model
        self._train(self.model)
        # save model
        if self.eval_coherence:
            self.eval_coherence_value()
        if self.autosave:
            self.save()

    def _train(self, model: Any) -> None:
        raise NotImplementedError

    def eval_coherence_value(
        self,
    ):
        mdl = self.model
        coh_metrics = {}
        for metric in self.coherence_metric_list:
            coh = tp.coherence.Coherence(mdl, coherence=metric)
            average_coherence = coh.get_score()
            coh_metrics[metric] = average_coherence
            coherence_per_topic = [coh.get_score(topic_id=k) for k in range(mdl.k)]
            if self.verbose:
                logger.info("==== Coherence : %s ====", metric)
                logger.info("Average: %s", average_coherence)
                logger.info("Per Topic: %s", coherence_per_topic)
        self._coherence_metrics_ = CoherenceMetrics(**coh_metrics)

    def save(self) -> None:
        self.save_model()
        self.save_train_summary()
        self.save_ll_per_words()
        self.plot_ll_per_words()
        self.save_dists_data()
        self.save_topic_top_words()
        self.generate_wordclouds()
        self.save_ldavis()
        self.save_model_summary()
        self.save_config()

    def save_model(self) -> None:
        self.model.save(self.model_file, full=self.save_full)
        logger.info("Model saved to %s", self.model_file)

    def save_ll_per_words(self) -> None:
        HyFI.save_dataframes(
            self.ll_per_words, self.ll_per_words_file, verbose=self.verbose
        )

    def plot_ll_per_words(self) -> None:
        df_ll = self.ll_per_words
        ax = df_ll.plot(x="iter", y="ll_per_word", kind="line")
        ax.set_xlabel("Iterations")
        ax.set_ylabel("Log-likelihood per word")
        ax.invert_yaxis()
        ax.get_figure().savefig(self.ll_per_words_fig_file, dpi=300, transparent=False)
        logger.info(
            "Log-likelihood per word plot saved to %s", self.ll_per_words_fig_file
        )

    def save_train_summary(self) -> None:
        coh_values = self.coherence_metrics_dict
        original_stdout = sys.stdout
        Path(self.train_summary_file).parent.mkdir(parents=True, exist_ok=True)
        with open(self.train_summary_file, "w") as f:
            sys.stdout = f  # Change the standard output to the file.
            self.model.summary(**self.train_summary_args.kwargs)
            if coh_values:
                print("<Topic Coherence Scores>")
                for cm, cv in coh_values.items():
                    print(f"| {cm}: {cv}")
            sys.stdout = original_stdout  # Reset the standard output.

    def save_model_summary(self) -> None:
        self._model_summary_ = ModelSummary(
            timestamp=self.timestamp,
            model_id=self.model_id,
            model_type=self.model_type,
            num_docs=self.num_docs,
            num_words=self.num_words,
            num_total_vocabs=self.num_total_vocabs,
            num_used_vocabs=self.num_used_vocab,
            seed=self.seed,
            model_args=self.model_args_dict,
            train_args=self.train_args_dict,
            perplexity=self.model.perplexity,
            coherence=self.coherence_metrics_dict,
        )
        if not self.model_summary_dict:
            logger.warning("Model summary is not available.")
        HyFI.append_to_jsonl(
            self.model_summary_dict,
            self.batch_model_summary_file,
        )
        logger.info("Model summary saved to %s", self.batch_model_summary_file)

    def save_dists_data(self):
        doc_topic_dists_df = self.get_doc_topic_dists_df(
            self.doc_topic_dists, self.doc_ids
        )
        if self.verbose:
            logger.info("==== Document-Topic Distributions ====")
            logger.info(doc_topic_dists_df.tail())
        HyFI.save_dataframes(
            doc_topic_dists_df,
            self.doc_topic_dists_file,
            verbose=self.verbose,
        )
        if self.verbose:
            logger.info("==== Topic-Word Distributions ====")
            logger.info(self.topic_term_dists_df.tail())
        HyFI.save_dataframes(
            self.topic_term_dists_df,
            self.topic_term_dists_file,
            verbose=self.verbose,
        )
        HyFI.save_wordlist(
            self.used_vocab,
            self.used_vocab_file,
            verbose=self.verbose,
        )

    def load(
        self,
        batch_name: Optional[str] = None,
        batch_num: Optional[int] = None,
        filepath: Optional[Union[str, Path]] = None,
        **config_kwargs,
    ):
        super().load_config(
            batch_name=batch_name,
            batch_num=batch_num,
            filepath=filepath,
            **config_kwargs,
        )
        self._load_model()
        self._load_ll_per_words()
        self._load_dists_data()

    def _load_ll_per_words(self):
        ll_df = HyFI.load_dataframes(self.ll_per_words_file, verbose=self.verbose)
        self._ll_per_words_ = [(ll.iter, ll.ll_per_word) for ll in ll_df.itertuples()]

    def _load_dists_data(self):
        self._doc_topic_dists_df_ = HyFI.load_dataframes(
            self.doc_topic_dists_file, verbose=self.verbose
        )
        self._doc_ids_ = self._doc_topic_dists_df_["id"].values.tolist()
        self._topic_term_dists_df_ = HyFI.load_dataframes(
            self.topic_term_dists_file, verbose=self.verbose
        )

    def save_ldavis(self):
        try:
            import pyLDAvis  # type: ignore[reportMissingImports]
        except ImportError:
            logger.warning(
                "pyLDAvis is not installed. Please install it to save LDAvis."
            )
            return

        prepared_data = pyLDAvis.prepare(
            topic_term_dists=self.topic_term_dists,
            doc_topic_dists=self.doc_topic_dists,
            doc_lengths=self.doc_num_words,
            vocab=self.used_vocab,
            term_frequency=self.term_frequency,
            start_index=0,
            sort_topics=False,
        )
        pyLDAvis.save_html(prepared_data, self.ldavis_file)
        logger.info("LDAvis saved to %s", self.ldavis_file)

    def get_topic_words(
        self,
        topic_id: int,
        top_n: int = 10,
    ) -> Dict[str, float]:
        return dict(self.model.get_topic_words(topic_id, top_n=top_n))

    @property
    def topic_top_words_file(self) -> str:
        f_ = f"{self.model_id}-topic_top_words.txt"
        return str(self.output_dir / f_)

    @property
    def topic_top_words_dists_file(self) -> str:
        f_ = f"{self.model_id}-topic_top_words_dists.csv"
        return str(self.output_dir / f_)

    def save_topic_top_words(self, top_n: int = 50):
        # set of top words
        topic_top_words = []
        # tuple of (topic_id, word, freq) for each topic
        topic_top_words_dists = []
        for topic_id in range(self.num_topics):
            topic_words = self.get_topic_words(topic_id, top_n=top_n)
            topic_top_words.extend(topic_words.keys())
            topic_words_freq_tuple = [
                (topic_id, w, topic_words[w]) for w in topic_words
            ]
            topic_top_words_dists.extend(topic_words_freq_tuple)

        HyFI.save_wordlist(
            list(set(topic_top_words)),
            self.topic_top_words_file,
            verbose=self.verbose,
        )
        HyFI.save_dataframes(
            pd.DataFrame(topic_top_words_dists, columns=["topic_id", "word", "freq"]),
            self.topic_top_words_dists_file,
            verbose=self.verbose,
        )

    def generate_wordclouds(
        self,
    ):
        wc_args = self.wc_args
        wc = wc_args.wc
        images = []
        for topic_id in range(self.num_topics):
            output_file = self.topic_wordcloud_file_format.format(topic_id=topic_id)
            img = wc.generate_from_frequencies(
                self.get_topic_words(topic_id, top_n=wc_args.top_n),
                output_file=output_file,
                verbose=self.verbose,
            )
            images.append(img)

        if wc_args.make_collage:
            titles = wc_args.titles or [f"Topic {i}" for i in range(self.num_topics)]
            logger.info("Making wordcloud collage with titles: %s", titles)
            output_dir = self.output_dir / "wordcloud_collage"
            output_file_format = self.model_id + "_wordcloud_{page_num:02d}.png"
            HyFI.make_subplot_pages_from_images(
                images,
                num_images_per_page=wc_args.num_images_per_page,
                num_cols=wc_args.num_cols,
                num_rows=wc_args.num_rows,
                output_dir=output_dir,
                output_file_format=output_file_format,
                titles=titles,
                title_fontsize=wc_args.title_fontsize,
                title_color=wc_args.title_color,
                figsize=wc_args.figsize,
                width_multiple=wc_args.width_multiple,
                height_multiple=wc_args.height_multiple,
                dpi=wc_args.dpi,
                verbose=self.verbose,
            )

    @property
    def inferred_doc_topic_dists_filename(self) -> str:
        return f"{self.model_id}-inferred_doc_topic_dists.parquet"

    def infer(
        self,
        corpus: Corpus,
        output_file: Optional[Union[str, Path]] = None,
        iterations: int = 100,
        tolerance: float = -1,
        num_workers: int = 0,
        together: bool = False,
    ):
        inferred_corpus, ll = self.model.infer(
            corpus.corpus,
            iter=iterations,
            tolerance=tolerance,
            workers=num_workers,
            together=together,
        )
        logger.info("Number of documents inferred: %d", len(inferred_corpus))
        output_file = output_file or (
            self.output_dir / "inferred_topics" / self.inferred_doc_topic_dists_filename
        )

        doc_ids = corpus.doc_ids
        doc_topic_dists = np.stack([doc.get_topic_dist() for doc in inferred_corpus])
        doc_topic_dists /= doc_topic_dists.sum(axis=1, keepdims=True)

        doc_topic_dists_df = self.get_doc_topic_dists_df(doc_topic_dists, doc_ids)
        ll_df = pd.DataFrame({"log_likelihood": ll})
        doc_topic_dists_df = pd.concat([doc_topic_dists_df, ll_df], axis=1)
        if self.verbose:
            logger.info("Inferred topics:\n%s", doc_topic_dists_df.head())
        HyFI.save_dataframes(
            doc_topic_dists_df,
            output_file,
            verbose=self.verbose,
        )
        logger.info("Inferred topics saved to %s", output_file)



from typing import List, Optional, Tuple

from hyfi.composer import BaseModel, Field


class NgramConfig(BaseModel):
    _config_group_ = "/ngrams"
    _config_name_ = "tp_ngrams"

    min_cf: int = Field(
        20, description="Minimum collection frequency of n-grams to be extracted."
    )
    min_df: int = Field(
        10, description="Minimum document frequency of n-grams to be extracted."
    )
    max_len: int = Field(5, description="Maximum length of n-grams to be extracted.")
    max_cand: int = Field(
        5000, description="Maximum number of n-grams to be extracted."
    )
    min_score: Optional[float] = Field(
        0.5, description="Minium PMI score of n-grams to be extracted."
    )
    normalized: bool = Field(
        True, description="Whether to normalize the PMI score of n-grams."
    )
    workers: int = Field(
        0, description="Number of workers to use for n-gram extraction."
    )
    delimiter: str = Field("_", description="Delimiter to use to concatenate n-grams.")
